/*  signals.c: signal handling */

/*  This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License , or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

    You may contact the author by:
       e-mail:  hlub@knoware.nl
*/

#include "rlwrap.h"

int command_is_dead = FALSE;
int commands_exit_status = 0;
int filters_exit_status = 0;
int deferred_adapt_commands_window_size = FALSE; /* whether we have to adapt clients winsize when accepting a line */
int signal_handlers_were_installed = FALSE; 
int received_sigALRM =  FALSE;

static void change_signalmask(int, int *);
static void child_died(int);
static void pass_on_signal(int);
static void handle_sigTSTP(int);



int adapt_tty_winsize(int, int);
static void wipe_textarea(struct winsize *old_size);


static int signals_to_be_passed_on[] = { SIGHUP, SIGINT,
  SIGQUIT, SIGABRT, SIGTERM, SIGCONT, SIGUSR1,
  SIGUSR2, SIGWINCH, 0
};

#ifdef DEBUG
static char *signal_names[] = { "SIGHUP", "SIGINT",
  "SIGQUIT", "SIGABRT", "SIGTERM", "SIGCONT", "SIGUSR1",
  "SIGUSR2", "SIGWINCH"
};				/* the *names* are inherently non-portable: signal numbers may
				   be different on different systems  */
                                /* @@@ We could have configure handle this! */

static void log_named_signal(int);
#else
static void handle_segfault(int);
#endif



int sigterm_received = FALSE;


void
mysignal(int sig, sighandler_type handler) {
#ifdef HAVE_SIGACTION
  struct sigaction action;
  action.sa_handler = handler;
  action.sa_flags = 0; /* no SA_RESTART */
  sigaction(sig, &action, NULL);
# else /* rlwrap running in Ye Olde Computer Museum?? */
  signal(sig, handler);
# endif
}


void
install_signal_handlers()
{
  int i;
  mysignal(SIGCHLD, &child_died);
  mysignal(SIGTSTP, &handle_sigTSTP);
#ifndef DEBUG
  mysignal(SIGSEGV, &handle_segfault);	/* we want core dumps when debugging, no polite excuses! */
#endif
  mysignal(SIGALRM, &handle_sigALRM);
  for (i = 0; signals_to_be_passed_on[i]; i++)
    mysignal(signals_to_be_passed_on[i], &pass_on_signal);
  signal_handlers_were_installed = TRUE;
}


static void uninstall_signal_handlers() {
  int i;
  mysignal(SIGCHLD, SIG_DFL); 
  mysignal(SIGTSTP, SIG_DFL);
  mysignal(SIGSEGV, SIG_DFL);
  mysignal(SIGALRM, SIG_DFL);
  for (i = 0; signals_to_be_passed_on[i]; i++)
    mysignal(signals_to_be_passed_on[i], SIG_DFL);
}


void
ignore_sigchld()
{
  mysignal(SIGCHLD, &do_nothing);
}


/* we'll call signals whose handlers mess with readline's internal state "messy signals"
   at present SIGTSTP and SIGWINCH are considered "messy". The following 2 functions are used
   in main.c to block messy signals except when waiting for I/O */

void
block_signals(int *sigs)
{
  change_signalmask(SIG_BLOCK, sigs);
}



void
unblock_signals(int *sigs)
{
  change_signalmask(SIG_UNBLOCK, sigs);
} 


void
block_all_passed_on_signals(void)
{
  change_signalmask(SIG_BLOCK, signals_to_be_passed_on);
}


  

static void
change_signalmask(int how, int *sigs)
{				/* sigs should point to a *zero-terminated* list of signals */
  int i;
  sigset_t mask;

  sigemptyset(&mask);
  for (i = 0; sigs[i]; i++)
    sigaddset(&mask, sigs[i]);
  sigprocmask(how, &mask, NULL);
}


void
unblock_all_signals()
{
  sigset_t mask;
  sigfillset(&mask);
  sigprocmask(SIG_UNBLOCK, &mask, NULL);
} 


void
block_all_signals()
{
  sigset_t mask;
  sigfillset(&mask); 
  sigprocmask(SIG_BLOCK, &mask, NULL);
}






static void
handle_sigTSTP(int signo)
{
  sigset_t mask;
  int error;

  DPRINTF0(DEBUG_SIGNALS, "got SIGTSTP");
  zero_select_timeout();
  if (command_pid && (error = kill(-command_pid, SIGTSTP))) {
    myerror("Failed to deliver signal");
  }

  if (within_line_edit)
    save_rl_state();

  /* unblock SIGTSTP, before we can send it to ourself (it is blocked while we're handling it) */

  sigemptyset(&mask);
  sigaddset(&mask, SIGTSTP);
  sigprocmask(SIG_UNBLOCK, &mask, NULL);
  mysignal(SIGTSTP, SIG_DFL);	/* reset disposition to default (i.e. suspend) */
  /* suspend myself (and supbprocess(es) in the same process group) */
  DPRINTF0(DEBUG_SIGNALS, "sending ourselves a SIGTSTP");
  kill(getpid(), SIGTSTP);

  /* keyboard gathers dust, kingdoms crumble,.... */

  /* Beautiful princess types "fg", (or her father tries to kill us...) and we wake up HERE: */

  DPRINTF0(DEBUG_SIGNALS, "woken up");

  mysignal(SIGTSTP, &handle_sigTSTP); /* because I had set it to SIG_DFL in order to suspend myself */ 
  /* we should really block SIGWINCH here ... */
  if (within_line_edit) {
    restore_rl_state();
    /* this is why we call SIGTSTP a "messy" signal */
  } else {
    set_echo(FALSE);
    cr();
    if (skip_rlwrap())
      return;
    move_cursor_to_start_of_prompt(ERASE);
    cook_prompt_if_necessary();
    my_putstr(saved_rl_state.cooked_prompt);
  }
  adapt_tty_winsize(STDIN_FILENO, master_pty_fd);	/* just in case */

}



/* signal handler, passes the signal to the child process group
   SIGWINCH is only passed after calling adapt_tty_winsize()
*/

static void
pass_on_signal(int signo)
{
  int ret, saved_errno = errno, pass_it_on = TRUE;
  zero_select_timeout();
#ifdef DEBUG
  log_named_signal(signo);
#endif
  
  if(pass_on_sigINT_as_sigTERM && signo == SIGINT)
    signo=SIGTERM;
  
  switch (signo) {
    case SIGWINCH:
      /* Make slave pty's winsize equal to that of STDIN. Pass the signal on *only if*
         winsize has changed. This is particularly important because we have the slave pty
         still open - when we pass on the signal the child will probably do a TIOCSWINSZ ioctl()
         on the slave pty  - causing us (the parent) to get a SIGWINCH again, thus getting stuck in a loop */
      pass_it_on = adapt_tty_winsize(STDIN_FILENO, master_pty_fd);
      break;
    case SIGTERM:
      sigterm_received = TRUE;
      break;
    default:
      break;
  }
  if (!command_pid)
    pass_it_on = FALSE;		/* no child (yet) to pass it on to */
  if (pass_it_on) {		/* we resend the signal to the process *group* of the child */
    ret = kill( -command_pid, signo);
    DPRINTF3(DEBUG_SIGNALS, "kill(%d,%d) = %d", -command_pid, signo, ret);
    we_just_got_a_signal = TRUE; /* signal to main loop that next command output should leave the prompt alone */ 
  }
  errno = saved_errno;
  
}



/* This function handles a terminal resize by copying from_fd's winsize
   to that of to_fd, keeping displayed line tidy, if possible
   (i.e. if terminal is capable enough, and we know enough about
   readlines internals).  A return value != 0 means that the size has
   changed.
*/

int
adapt_tty_winsize(int from_fd, int to_fd)
{
  int ret;

  struct winsize old_winsize = winsize;

  ret = ioctl(from_fd, TIOCGWINSZ, &winsize);
  DPRINTF1(DEBUG_SIGNALS, "ioctl (..., TIOCGWINSZ) = %d", ret);
  if (winsize.ws_col != old_winsize.ws_col || winsize.ws_row != old_winsize.ws_row) { 
    if (always_readline &&!dont_wrap_command_waits())                       /* if --always_readline option is set, the client will probably spew a */
      deferred_adapt_commands_window_size = TRUE;   /* volley of control chars at us when its terminal is resized. Hence we don't do it now */
    else {  
      ret = ioctl(to_fd, TIOCSWINSZ, &winsize); 
      DPRINTF1(DEBUG_SIGNALS, "ioctl (..., TIOCSWINSZ) = %d", ret);
    }	
    DPRINTF2(DEBUG_READLINE, "rl_prompt: %s, line_buffer: %s", mangle_string_for_debug_log(rl_prompt, 100), rl_line_buffer);
    rl_set_screen_size(winsize.ws_row, winsize.ws_col);	/* this is safe: we know that right now rlwrap is not calling
                                                                   the readline library because we keep SIGWINCH blocked all the time */

    if (!within_line_edit && !skip_rlwrap()) {
      wipe_textarea(&old_winsize);
     
      received_WINCH = TRUE;           /* we can't start line edit in signal handler, so we only set a flag */
    } else if (within_line_edit) {	/* try to keep displayed line tidy */
      wipe_textarea(&old_winsize);
      rl_on_new_line();
      rl_redisplay();
      
    }
    
    return (!always_readline || dont_wrap_command_waits()); /* pass the signal on (except when always_readline is set and command is not waiting) */
  } else {			/* window size has not changed */
    return FALSE;
  }
  
}

/* After a resize, clear all the lines that were occupied by prompt + line buffer before the resize */
static
void wipe_textarea(struct winsize *old_winsize)
{
  int point, lineheight, linelength, cursor_height, i, promptlength;
  if (!prompt_is_single_line()) {	/* Don't need to do anything in horizontal_scroll_mode  */
    promptlength = colourless_strlen((saved_rl_state.cooked_prompt ? saved_rl_state.cooked_prompt:  saved_rl_state.raw_prompt), NULL); 
    linelength = (within_line_edit ? strlen(rl_line_buffer) : 0) + promptlength;
    point = (within_line_edit ? rl_point : 0) + promptlength;
    lineheight = (linelength == 0 ? 0 : (1 + (max(point, (linelength - 1)) / old_winsize -> ws_col)));
    if (lineheight > 1 && term_cursor_up != NULL && term_cursor_down != NULL) {
      /* i.e. if we have multiple rows displayed, and we can clean them up first */
      cr(); 
      cursor_height = point / old_winsize -> ws_col;    /* cursor is still on old line */
      DPRINTF2(DEBUG_SIGNALS, "lineheight: %d, cursor_height: %d", lineheight, cursor_height);
      for (i = 1 + cursor_height; i < lineheight; i++)
	curs_down();	/* ...move it down to last line */
      for (i = lineheight; i > 1; i--) {	/* go up again, erasing every line */
	clear_line(); 
	curs_up();
      }
    }
    clear_line();
    cr();
  }
}	




static void
child_died(int unused)
{
  int saved_errno;

  zero_select_timeout();
  saved_errno = errno;
  DPRINTF0(DEBUG_SIGNALS, "Caught SIGCHLD");
  
  if(command_pid && waitpid(command_pid, &commands_exit_status, WNOHANG)) {
    DPRINTF2(DEBUG_SIGNALS, "child (pid %d) has died, exit status: %x", command_pid, commands_exit_status);
      command_is_dead = TRUE;
      command_pid = 0;		/* thus we know that there is no child anymore to pass signals to */
  } else if (filter_pid && waitpid(filter_pid, &filters_exit_status, WNOHANG)) { 
    DPRINTF2(DEBUG_SIGNALS, "filter (pid %d) has died, exit status: %x", filter_pid, filters_exit_status);
      filter_is_dead = TRUE;
      filter_pid = 0;
  } else  {
    return; /* ignore */
  }	

  errno = saved_errno;  
  return;			/* allow remaining output from child to be processed in main loop */
  /* (so that we will see childs good-bye talk)                     */
  /* this will then clean up and terminate                          */

}





#ifdef DEBUG
static void
log_named_signal(int signo)
{
  if (debug) {
    int *p, i;

    for (i = 0, p = signals_to_be_passed_on; *p; p++, i++) {
      if (*p == signo) {
	DPRINTF2(DEBUG_SIGNALS, "got signal %d (%s on most systems)", signo,
		 signal_names[i]);
	break;
      }
    }
    if (!*p)
      DPRINTF1(DEBUG_SIGNALS, "got signal %d", signo);
  }
}
#else
static void
handle_segfault(int sig)
{				/* Even after sudden and unexpected death, leave the terminal in a tidy state */

  printf("\nrlwrap: Oops, segfault -  this should not have happened!\n"
	 "If you need a core dump, re-configure with --enable-debug and rebuild\n"
	 "Resetting terminal and cleaning up...\n");
  if (colour_the_prompt)
    write(STDOUT_FILENO,"\033[0m",4); /* reset terminal colours */
  if (terminal_settings_saved)
    tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_terminal_settings);
 
  exit(EXIT_FAILURE);
}
#endif


static int coredump(int status) {
  #ifdef WCOREDUMP
    return WCOREDUMP(status);
  #else
    return 0;
  #endif
}

void suicide_by(int signal, int status) {
  uninstall_signal_handlers();
  if (signal == SIGSEGV || signal == SIGABRT || signal == SIGALRM) {
    errno = 0;
    mywarn("%s killed by signal %d%s - %s will now kill itself with the same signal",
	    command_name, signal, (coredump(status) ? " (core dumped)" : ""), program_name);
  }	
  unblock_all_signals();
  DPRINTF1(DEBUG_SIGNALS, "suicide by signal  %d", signal);
  kill(getpid(), signal);
  /* still alive?: */
  sleep(1);
  exit(0);
}

static int myalarm_was_set = FALSE;


/* drop-in replacement for alarm (with msec precision). Also sets global flag
   myalarm_was_set */

void myalarm(int msecs) {
#ifdef HAVE_SETITIMER
  struct itimerval awhile = {{0,0},{0,0}};
  awhile.it_value.tv_usec = msecs * 1000;
  received_sigALRM = FALSE;
  setitimer(ITIMER_REAL, &awhile, NULL);
#else
  received_sigALRM = FALSE;
  alarm(msecs == 0 ? 0 : 1 + msecs/1000)); 
#endif
  DPRINTF1(DEBUG_SIGNALS, "set alarm (%d msecs)", msecs);
  if (msecs == 0)
    return;
  myalarm_was_set = TRUE;
}

void
handle_sigALRM(int signo) {
  received_sigALRM = TRUE;
  assert(myalarm_was_set == TRUE); /* cry wolf if sigALRM is caught when none was requested by myalarm */
  myalarm_was_set= FALSE;
  DPRINTF0(DEBUG_SIGNALS, "got sigALRM");
}
